#include <OgreCamera.h>
#include <OgreLogManager.h>
#include "OgreSolidBoundingBox.h"
#include "OgrePlatformHierarchyInterface.h"
#include "OgrePlatformOcclusionQuery.h"

namespace Ogre {

//-----------------------------------------------------------------------
PlatformHierarchyInterface::PlatformHierarchyInterface(SceneManager *sm, RenderSystem *rsys):
mSceneManager(sm), mRenderSystem(rsys), mSolidBoundingBox(NULL),
mCamera(NULL), mCullCamera(NULL), mOnlyShadowCasters(false)
{
}
//-----------------------------------------------------------------------
void PlatformHierarchyInterface::CreateCulledNodeMaterials()
{
	// material for frustum culled nodes
	MaterialPtr mat = MaterialManager::getSingleton().getByName("FrustumCulledNodesMaterial");
	
	if (mat.isNull())
	{
		mat = MaterialManager::getSingleton().create("FrustumCulledNodesMaterial", 
			ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME);
		mat->createTechnique()->createPass();
		
		mat->getTechnique(0)->getPass(0)->setAmbient(1,0,0);
		mat->getTechnique(0)->getPass(0)->setLightingEnabled(true);
		//mat->getTechnique(0)->getPass(0)->setDepthCheckEnabled(false);
		
		mat->load();
	}

	// material for query culled nodes
	mat = MaterialManager::getSingleton().getByName("QueryCulledNodesMaterial");

	if (mat.isNull())
	{
		mat = MaterialManager::getSingleton().create("QueryCulledNodesMaterial", 
			ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME);
		mat->createTechnique()->createPass();
		
		mat->getTechnique(0)->getPass(0)->setAmbient(0,0,1);
		mat->getTechnique(0)->getPass(0)->setLightingEnabled(true);
		mat->load();
	}
}
//-----------------------------------------------------------------------
PlatformHierarchyInterface::~PlatformHierarchyInterface()
{
	DeleteQueries();

	if (mSolidBoundingBox)
		delete mSolidBoundingBox;
}
//-----------------------------------------------------------------------
void PlatformHierarchyInterface::DeleteQueries()
{
	for (int i=0; i < (int)mOcclusionQueries.size(); ++i)
		delete mOcclusionQueries[i];

	mOcclusionQueries.clear();
}
//-----------------------------------------------------------------------
void PlatformHierarchyInterface::RenderBoundingBox(AxisAlignedBox *box)
{
	static RenderOperation ro;

	SolidBoundingBox *solidBox = GetSolidBoundingBox();
	
	mRenderSystem->_setWorldMatrix(Ogre::Matrix4::IDENTITY);
	mSceneManager->useRenderableViewProjMode(solidBox);
	//mSceneManager->setOcclusionPass(solidBox->getTechnique()->getPass(0));
	SetOcclusionPass();

	solidBox->SetupBoundingBoxVertices(*box);
				
	solidBox->getRenderOperation(ro);
	ro.srcRenderable = solidBox;
	mRenderSystem->_render(ro);

	// Render two halfes of the bounding box (using triangle fans)
	/*for(int halfIdx = 0; halfIdx < 2; ++halfIdx)
	{
		solidBox->SetupBoundingBoxVertices(*box, halfIdx == 1);
				
		solidBox->getRenderOperation(ro);
		ro.srcRenderable = solidBox;
		mRenderSystem->_render(ro);
	}*/
}
//-----------------------------------------------------------------------
void PlatformHierarchyInterface::SetCamera(Ogre::Camera *cam)
{
	mCamera = cam;
}
//-----------------------------------------------------------------------
void PlatformHierarchyInterface::SetCullCamera(Ogre::Camera *cullCam)
{
	mCullCamera = cullCam;
}
//-----------------------------------------------------------------------
GtpVisibility::OcclusionQuery *PlatformHierarchyInterface::GetNextOcclusionQuery()
{
	if (mCurrentTestIdx == mOcclusionQueries.size())
	{
		mOcclusionQueries.push_back(new PlatformOcclusionQuery(mRenderSystem));
	}
	
	return mOcclusionQueries[mCurrentTestIdx ++];
}
//-----------------------------------------------------------------------
void PlatformHierarchyInterface::InitFrame(GtpVisibility::HierarchyNode *root, 
										   Camera *cam, Camera *cullCam)
{
	GtpVisibility::HierarchyInterface::InitFrame(root);
	mPreviousNode = NULL;

	SetCamera(cam);

	if (cullCam)
	{
		SetCullCamera(cullCam);
	}
	else
	{
		SetCullCamera(cam);
	}

	CreateCulledNodeMaterials();
}
//-----------------------------------------------------------------------
void PlatformHierarchyInterface::SetSceneManager(SceneManager *sm)
{
	mSceneManager = sm;
}
//-----------------------------------------------------------------------
void PlatformHierarchyInterface::SetRenderSystem(RenderSystem *rsys)
{
	mRenderSystem = rsys;
}
//-----------------------------------------------------------------------
bool PlatformHierarchyInterface::CheckFrustumVisible(GtpVisibility::HierarchyNode *node, 
													 bool &intersects)
{
#ifdef GTP_VISIBILITY_MODIFIED_OGRE
	return mCullCamera->isVisible(*GetBoundingBox(node), intersects);
#else
	return true;
#endif
}
//-----------------------------------------------------------------------
GtpVisibility::OcclusionQuery *PlatformHierarchyInterface::IssueOcclusionQuery(
	GtpVisibility::HierarchyNode *node, const bool wasVisible) 
{
	// get next available test id
	GtpVisibility::OcclusionQuery *query = GetNextOcclusionQuery();

	//-- the actual query test
	query->BeginQuery();
			
	// if node is leaf and was visible => will be rendered anyway.
	// In this case we can also test with the real geometry.
	// if camera for culling is different from camera for rendering => cannot optimize
	if (mUseOptimization && wasVisible && IsLeaf(node) && (mCamera == mCullCamera))
	{
		//LogManager::getSingleton().logMessage("render node\n");
		RenderNode(node);
	}
	else
	{
		//LogManager::getSingleton().logMessage("render box\n");
		RenderBoundingBox(GetBoundingBox(node));
	}

	query->EndQuery();
	
	return query;
}
//-----------------------------------------------------------------------
void PlatformHierarchyInterface::SetOcclusionPass()
{
    // disable vertex and fragment program
	mRenderSystem->unbindGpuProgram(GPT_VERTEX_PROGRAM);
	mRenderSystem->unbindGpuProgram(GPT_FRAGMENT_PROGRAM);
	
	// disable lighting
	mRenderSystem->setLightingEnabled(false);

    // Disable remaining texture units
    mRenderSystem->_disableTextureUnitsFrom(0);

    //--Set up non-texture related material settings
   
	// Depth buffer settings
	mRenderSystem->_setDepthBufferParams(true, false, CMPF_LESS_EQUAL);
	// Set colour write mode off
	mRenderSystem->_setColourBufferWriteEnabled(false, false, false, false);
}
//-----------------------------------------------------------------------
SolidBoundingBox *PlatformHierarchyInterface::GetSolidBoundingBox()
{
	if (!mSolidBoundingBox)
		mSolidBoundingBox = new SolidBoundingBox;

	return mSolidBoundingBox;
}
//-----------------------------------------------------------------------
void PlatformHierarchyInterface::SetOnlyShadowCasters(bool onlyShadowCasters)
{
	mOnlyShadowCasters = onlyShadowCasters;
	//mIsShadowPass = mOnlyShadowCasters ;
}
//-----------------------------------------------------------------------
bool PlatformHierarchyInterface::GetOnlyShadowCasters()
{
	return mOnlyShadowCasters;
}
} // namespace Ogre